# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

r_wrapper_poismf <- function(A, B, dimA, dimB, k, Xr, Xr_ind_int, Xr_indptr_int, Xc, Xc_ind_int, Xc_indptr_int, nnz, l1_reg, l2_reg, niter, npass, step_size, use_cg, nthreads) {
    invisible(.Call(`_poismf_r_wrapper_poismf`, A, B, dimA, dimB, k, Xr, Xr_ind_int, Xr_indptr_int, Xc, Xc_ind_int, Xc_indptr_int, nnz, l1_reg, l2_reg, niter, npass, step_size, use_cg, nthreads))
}

predict_multiple <- function(A, B, k, npred, ia, ib, out, nthreads) {
    invisible(.Call(`_poismf_predict_multiple`, A, B, k, npred, ia, ib, out, nthreads))
}

calc_fun_single_R <- function(x_R, X_R, X_ind, nnz_this, F_R, Fsum, n, l2_reg, grad) {
    .Call(`_poismf_calc_fun_single_R`, x_R, X_R, X_ind, nnz_this, F_R, Fsum, n, l2_reg, grad)
}

calc_grad_single_R <- function(x_R, X_R, X_ind, nnz_this, F_R, Fsum, n, l2_reg, grad_R) {
    .Call(`_poismf_calc_grad_single_R`, x_R, X_R, X_ind, nnz_this, F_R, Fsum, n, l2_reg, grad_R)
}

select_topN <- function(preds, ix, topN) {
    invisible(.Call(`_poismf_select_topN`, preds, ix, topN))
}

